@@ -1,18 +1,35 @@
-#line 1 "parser.rl"
-/* vim: set cin et sw=4 ts=4: */
 
+#line 1 "parser.rl"
 #include "ruby.h"
-#include "re.h"
-#include "st.h"
 #include "unicode.h"
+#if HAVE_RE_H
+#include "re.h"
+#endif
+#if HAVE_RUBY_ST_H
+#include "ruby/st.h"
+#endif
+#if HAVE_ST_H
+#include "st.h"
+#endif
 
 #define EVIL 0x666
 
+#ifndef RHASH_TBL
+#define RHASH_TBL(hsh) (RHASH(hsh)->tbl)
+#endif
+
+#ifdef HAVE_RUBY_ENCODING_H
+#include "ruby/encoding.h"
+#define FORCE_UTF8(obj) rb_enc_associate((obj), rb_utf8_encoding())
+#else
+#define FORCE_UTF8(obj)
+#endif
+
 static VALUE mJSON, mExt, cParser, eParserError, eNestingError;
 static VALUE CNaN, CInfinity, CMinusInfinity;
 
 static ID i_json_creatable_p, i_json_create, i_create_id, i_create_additions,
-          i_chr, i_max_nesting, i_allow_nan; 
+          i_chr, i_max_nesting, i_allow_nan, i_object_class, i_array_class; 
 
 #define MinusInfinity "-Infinity"
 
@@ -25,6 +42,8 @@ typedef struct JSON_ParserStruct {
     int max_nesting;
     int current_nesting;
     int allow_nan;
+    VALUE object_class;
+    VALUE array_class;
 } JSON_Parser;
 
 static char *JSON_parse_object(JSON_Parser *json, char *p, char *pe, VALUE *result);
@@ -38,42 +57,46 @@ static char *JSON_parse_float(JSON_Parser *json, char *p, char *pe, VALUE *resul
     JSON_Parser *json;                      \
     Data_Get_Struct(self, JSON_Parser, json);
 
-#line 66 "parser.rl"
+
+#line 84 "parser.rl"
 
 
 
-#line 46 "parser.c"
+#line 66 "parser.c"
 static const int JSON_object_start = 1;
 static const int JSON_object_first_final = 27;
 static const int JSON_object_error = 0;
 
 static const int JSON_object_en_main = 1;
 
-#line 99 "parser.rl"
+
+#line 117 "parser.rl"
 
 
 static char *JSON_parse_object(JSON_Parser *json, char *p, char *pe, VALUE *result)
 {
     int cs = EVIL;
     VALUE last_name = Qnil;
+    VALUE object_class = json->object_class;
 
     if (json->max_nesting && json->current_nesting > json->max_nesting) {
-        rb_raise(eNestingError, "nesting of %d is to deep", json->current_nesting);
+        rb_raise(eNestingError, "nesting of %d is too deep", json->current_nesting);
     }
 
-    *result = rb_hash_new();
+    *result = NIL_P(object_class) ? rb_hash_new() : rb_class_new_instance(0, 0, object_class);
 
     
-#line 68 "parser.c"
+#line 90 "parser.c"
 	{
 	cs = JSON_object_start;
 	}
-#line 113 "parser.rl"
+
+#line 132 "parser.rl"
     
-#line 74 "parser.c"
+#line 97 "parser.c"
 	{
 	if ( p == pe )
-		goto _out;
+		goto _test_eof;
 	switch ( cs )
 	{
 case 1:
@@ -81,10 +104,11 @@ case 1:
 		goto st2;
 	goto st0;
 st0:
-	goto _out0;
+cs = 0;
+	goto _out;
 st2:
 	if ( ++p == pe )
-		goto _out2;
+		goto _test_eof2;
 case 2:
 	switch( (*p) ) {
 		case 13: goto st2;
@@ -97,17 +121,17 @@ case 2:
 		goto st2;
 	goto st0;
 tr2:
-#line 85 "parser.rl"
+#line 103 "parser.rl"
 	{
         char *np = JSON_parse_string(json, p, pe, &last_name);
-        if (np == NULL) goto _out3; else {p = (( np))-1;}
+        if (np == NULL) { p--; {p++; cs = 3; goto _out;} } else {p = (( np))-1;}
     }
 	goto st3;
 st3:
 	if ( ++p == pe )
-		goto _out3;
+		goto _test_eof3;
 case 3:
-#line 111 "parser.c"
+#line 135 "parser.c"
 	switch( (*p) ) {
 		case 13: goto st3;
 		case 32: goto st3;
@@ -119,7 +143,7 @@ case 3:
 	goto st0;
 st4:
 	if ( ++p == pe )
-		goto _out4;
+		goto _test_eof4;
 case 4:
 	switch( (*p) ) {
 		case 42: goto st5;
@@ -128,14 +152,14 @@ case 4:
 	goto st0;
 st5:
 	if ( ++p == pe )
-		goto _out5;
+		goto _test_eof5;
 case 5:
 	if ( (*p) == 42 )
 		goto st6;
 	goto st5;
 st6:
 	if ( ++p == pe )
-		goto _out6;
+		goto _test_eof6;
 case 6:
 	switch( (*p) ) {
 		case 42: goto st6;
@@ -144,14 +168,14 @@ case 6:
 	goto st5;
 st7:
 	if ( ++p == pe )
-		goto _out7;
+		goto _test_eof7;
 case 7:
 	if ( (*p) == 10 )
 		goto st3;
 	goto st7;
 st8:
 	if ( ++p == pe )
-		goto _out8;
+		goto _test_eof8;
 case 8:
 	switch( (*p) ) {
 		case 13: goto st8;
@@ -174,12 +198,12 @@ case 8:
 		goto st8;
 	goto st0;
 tr11:
-#line 74 "parser.rl"
+#line 92 "parser.rl"
 	{
         VALUE v = Qnil;
         char *np = JSON_parse_value(json, p, pe, &v); 
         if (np == NULL) {
-            goto _out9;
+            p--; {p++; cs = 9; goto _out;}
         } else {
             rb_hash_aset(*result, last_name, v);
             {p = (( np))-1;}
@@ -188,9 +212,9 @@ tr11:
 	goto st9;
 st9:
 	if ( ++p == pe )
-		goto _out9;
+		goto _test_eof9;
 case 9:
-#line 194 "parser.c"
+#line 218 "parser.c"
 	switch( (*p) ) {
 		case 13: goto st9;
 		case 32: goto st9;
@@ -203,7 +227,7 @@ case 9:
 	goto st0;
 st10:
 	if ( ++p == pe )
-		goto _out10;
+		goto _test_eof10;
 case 10:
 	switch( (*p) ) {
 		case 13: goto st10;
@@ -216,7 +240,7 @@ case 10:
 	goto st0;
 st11:
 	if ( ++p == pe )
-		goto _out11;
+		goto _test_eof11;
 case 11:
 	switch( (*p) ) {
 		case 42: goto st12;
@@ -225,14 +249,14 @@ case 11:
 	goto st0;
 st12:
 	if ( ++p == pe )
-		goto _out12;
+		goto _test_eof12;
 case 12:
 	if ( (*p) == 42 )
 		goto st13;
 	goto st12;
 st13:
 	if ( ++p == pe )
-		goto _out13;
+		goto _test_eof13;
 case 13:
 	switch( (*p) ) {
 		case 42: goto st13;
@@ -241,14 +265,14 @@ case 13:
 	goto st12;
 st14:
 	if ( ++p == pe )
-		goto _out14;
+		goto _test_eof14;
 case 14:
 	if ( (*p) == 10 )
 		goto st10;
 	goto st14;
 st15:
 	if ( ++p == pe )
-		goto _out15;
+		goto _test_eof15;
 case 15:
 	switch( (*p) ) {
 		case 42: goto st16;
@@ -257,14 +281,14 @@ case 15:
 	goto st0;
 st16:
 	if ( ++p == pe )
-		goto _out16;
+		goto _test_eof16;
 case 16:
 	if ( (*p) == 42 )
 		goto st17;
 	goto st16;
 st17:
 	if ( ++p == pe )
-		goto _out17;
+		goto _test_eof17;
 case 17:
 	switch( (*p) ) {
 		case 42: goto st17;
@@ -273,24 +297,24 @@ case 17:
 	goto st16;
 st18:
 	if ( ++p == pe )
-		goto _out18;
+		goto _test_eof18;
 case 18:
 	if ( (*p) == 10 )
 		goto st9;
 	goto st18;
 tr4:
-#line 90 "parser.rl"
-	{ goto _out27; }
+#line 108 "parser.rl"
+	{ p--; {p++; cs = 27; goto _out;} }
 	goto st27;
 st27:
 	if ( ++p == pe )
-		goto _out27;
+		goto _test_eof27;
 case 27:
-#line 290 "parser.c"
+#line 314 "parser.c"
 	goto st0;
 st19:
 	if ( ++p == pe )
-		goto _out19;
+		goto _test_eof19;
 case 19:
 	switch( (*p) ) {
 		case 42: goto st20;
@@ -299,14 +323,14 @@ case 19:
 	goto st0;
 st20:
 	if ( ++p == pe )
-		goto _out20;
+		goto _test_eof20;
 case 20:
 	if ( (*p) == 42 )
 		goto st21;
 	goto st20;
 st21:
 	if ( ++p == pe )
-		goto _out21;
+		goto _test_eof21;
 case 21:
 	switch( (*p) ) {
 		case 42: goto st21;
@@ -315,14 +339,14 @@ case 21:
 	goto st20;
 st22:
 	if ( ++p == pe )
-		goto _out22;
+		goto _test_eof22;
 case 22:
 	if ( (*p) == 10 )
 		goto st8;
 	goto st22;
 st23:
 	if ( ++p == pe )
-		goto _out23;
+		goto _test_eof23;
 case 23:
 	switch( (*p) ) {
 		case 42: goto st24;
@@ -331,14 +355,14 @@ case 23:
 	goto st0;
 st24:
 	if ( ++p == pe )
-		goto _out24;
+		goto _test_eof24;
 case 24:
 	if ( (*p) == 42 )
 		goto st25;
 	goto st24;
 st25:
 	if ( ++p == pe )
-		goto _out25;
+		goto _test_eof25;
 case 25:
 	switch( (*p) ) {
 		case 42: goto st25;
@@ -347,43 +371,44 @@ case 25:
 	goto st24;
 st26:
 	if ( ++p == pe )
-		goto _out26;
+		goto _test_eof26;
 case 26:
 	if ( (*p) == 10 )
 		goto st2;
 	goto st26;
 	}
-	_out0: cs = 0; goto _out; 
-	_out2: cs = 2; goto _out; 
-	_out3: cs = 3; goto _out; 
-	_out4: cs = 4; goto _out; 
-	_out5: cs = 5; goto _out; 
-	_out6: cs = 6; goto _out; 
-	_out7: cs = 7; goto _out; 
-	_out8: cs = 8; goto _out; 
-	_out9: cs = 9; goto _out; 
-	_out10: cs = 10; goto _out; 
-	_out11: cs = 11; goto _out; 
-	_out12: cs = 12; goto _out; 
-	_out13: cs = 13; goto _out; 
-	_out14: cs = 14; goto _out; 
-	_out15: cs = 15; goto _out; 
-	_out16: cs = 16; goto _out; 
-	_out17: cs = 17; goto _out; 
-	_out18: cs = 18; goto _out; 
-	_out27: cs = 27; goto _out; 
-	_out19: cs = 19; goto _out; 
-	_out20: cs = 20; goto _out; 
-	_out21: cs = 21; goto _out; 
-	_out22: cs = 22; goto _out; 
-	_out23: cs = 23; goto _out; 
-	_out24: cs = 24; goto _out; 
-	_out25: cs = 25; goto _out; 
-	_out26: cs = 26; goto _out; 
+	_test_eof2: cs = 2; goto _test_eof; 
+	_test_eof3: cs = 3; goto _test_eof; 
+	_test_eof4: cs = 4; goto _test_eof; 
+	_test_eof5: cs = 5; goto _test_eof; 
+	_test_eof6: cs = 6; goto _test_eof; 
+	_test_eof7: cs = 7; goto _test_eof; 
+	_test_eof8: cs = 8; goto _test_eof; 
+	_test_eof9: cs = 9; goto _test_eof; 
+	_test_eof10: cs = 10; goto _test_eof; 
+	_test_eof11: cs = 11; goto _test_eof; 
+	_test_eof12: cs = 12; goto _test_eof; 
+	_test_eof13: cs = 13; goto _test_eof; 
+	_test_eof14: cs = 14; goto _test_eof; 
+	_test_eof15: cs = 15; goto _test_eof; 
+	_test_eof16: cs = 16; goto _test_eof; 
+	_test_eof17: cs = 17; goto _test_eof; 
+	_test_eof18: cs = 18; goto _test_eof; 
+	_test_eof27: cs = 27; goto _test_eof; 
+	_test_eof19: cs = 19; goto _test_eof; 
+	_test_eof20: cs = 20; goto _test_eof; 
+	_test_eof21: cs = 21; goto _test_eof; 
+	_test_eof22: cs = 22; goto _test_eof; 
+	_test_eof23: cs = 23; goto _test_eof; 
+	_test_eof24: cs = 24; goto _test_eof; 
+	_test_eof25: cs = 25; goto _test_eof; 
+	_test_eof26: cs = 26; goto _test_eof; 
 
+	_test_eof: {}
 	_out: {}
 	}
-#line 114 "parser.rl"
+
+#line 133 "parser.rl"
 
     if (cs >= JSON_object_first_final) {
         if (RTEST(json->create_id)) {
@@ -402,14 +427,15 @@ case 26:
 }
 
 
-#line 406 "parser.c"
+#line 431 "parser.c"
 static const int JSON_value_start = 1;
 static const int JSON_value_first_final = 21;
 static const int JSON_value_error = 0;
 
 static const int JSON_value_en_main = 1;
 
-#line 212 "parser.rl"
+
+#line 231 "parser.rl"
 
 
 static char *JSON_parse_value(JSON_Parser *json, char *p, char *pe, VALUE *result)
@@ -417,16 +443,17 @@ static char *JSON_parse_value(JSON_Parser *json, char *p, char *pe, VALUE *resul
     int cs = EVIL;
 
     
-#line 421 "parser.c"
+#line 447 "parser.c"
 	{
 	cs = JSON_value_start;
 	}
-#line 219 "parser.rl"
+
+#line 238 "parser.rl"
     
-#line 427 "parser.c"
+#line 454 "parser.c"
 	{
 	if ( p == pe )
-		goto _out;
+		goto _test_eof;
 	switch ( cs )
 	{
 case 1:
@@ -445,259 +472,261 @@ case 1:
 		goto tr2;
 	goto st0;
 st0:
-	goto _out0;
+cs = 0;
+	goto _out;
 tr0:
-#line 160 "parser.rl"
+#line 179 "parser.rl"
 	{
         char *np = JSON_parse_string(json, p, pe, result);
-        if (np == NULL) goto _out21; else {p = (( np))-1;}
+        if (np == NULL) { p--; {p++; cs = 21; goto _out;} } else {p = (( np))-1;}
     }
 	goto st21;
 tr2:
-#line 165 "parser.rl"
+#line 184 "parser.rl"
 	{
         char *np;
         if(pe > p + 9 && !strncmp(MinusInfinity, p, 9)) {
             if (json->allow_nan) {
                 *result = CMinusInfinity;
                 {p = (( p + 10))-1;}
-                goto _out21;
+                p--; {p++; cs = 21; goto _out;}
             } else {
-                rb_raise(eParserError, "unexpected token at '%s'", p);
+                rb_raise(eParserError, "%u: unexpected token at '%s'", __LINE__, p);
             }
         }
         np = JSON_parse_float(json, p, pe, result);
         if (np != NULL) {p = (( np))-1;}
         np = JSON_parse_integer(json, p, pe, result);
         if (np != NULL) {p = (( np))-1;}
-        goto _out21;
+        p--; {p++; cs = 21; goto _out;}
     }
 	goto st21;
 tr5:
-#line 183 "parser.rl"
+#line 202 "parser.rl"
 	{ 
         char *np;
-        json->current_nesting += 1;
+        json->current_nesting++;
         np = JSON_parse_array(json, p, pe, result);
-        json->current_nesting -= 1;
-        if (np == NULL) goto _out21; else {p = (( np))-1;}
+        json->current_nesting--;
+        if (np == NULL) { p--; {p++; cs = 21; goto _out;} } else {p = (( np))-1;}
     }
 	goto st21;
 tr9:
-#line 191 "parser.rl"
+#line 210 "parser.rl"
 	{ 
         char *np;
-        json->current_nesting += 1;
+        json->current_nesting++;
         np =  JSON_parse_object(json, p, pe, result);
-        json->current_nesting -= 1;
-        if (np == NULL) goto _out21; else {p = (( np))-1;}
+        json->current_nesting--;
+        if (np == NULL) { p--; {p++; cs = 21; goto _out;} } else {p = (( np))-1;}
     }
 	goto st21;
 tr16:
-#line 153 "parser.rl"
+#line 172 "parser.rl"
 	{
         if (json->allow_nan) {
             *result = CInfinity;
         } else {
-            rb_raise(eParserError, "unexpected token at '%s'", p - 8);
+            rb_raise(eParserError, "%u: unexpected token at '%s'", __LINE__, p - 8);
         }
     }
 	goto st21;
 tr18:
-#line 146 "parser.rl"
+#line 165 "parser.rl"
 	{
         if (json->allow_nan) {
             *result = CNaN;
         } else {
-            rb_raise(eParserError, "unexpected token at '%s'", p - 2);
+            rb_raise(eParserError, "%u: unexpected token at '%s'", __LINE__, p - 2);
         }
     }
 	goto st21;
 tr22:
-#line 140 "parser.rl"
+#line 159 "parser.rl"
 	{
         *result = Qfalse;
     }
 	goto st21;
 tr25:
-#line 137 "parser.rl"
+#line 156 "parser.rl"
 	{
         *result = Qnil;
     }
 	goto st21;
 tr28:
-#line 143 "parser.rl"
+#line 162 "parser.rl"
 	{
         *result = Qtrue;
     }
 	goto st21;
 st21:
 	if ( ++p == pe )
-		goto _out21;
+		goto _test_eof21;
 case 21:
-#line 199 "parser.rl"
-	{ goto _out21; }
-#line 541 "parser.c"
+#line 218 "parser.rl"
+	{ p--; {p++; cs = 21; goto _out;} }
+#line 569 "parser.c"
 	goto st0;
 st2:
 	if ( ++p == pe )
-		goto _out2;
+		goto _test_eof2;
 case 2:
 	if ( (*p) == 110 )
 		goto st3;
 	goto st0;
 st3:
 	if ( ++p == pe )
-		goto _out3;
+		goto _test_eof3;
 case 3:
 	if ( (*p) == 102 )
 		goto st4;
 	goto st0;
 st4:
 	if ( ++p == pe )
-		goto _out4;
+		goto _test_eof4;
 case 4:
 	if ( (*p) == 105 )
 		goto st5;
 	goto st0;
 st5:
 	if ( ++p == pe )
-		goto _out5;
+		goto _test_eof5;
 case 5:
 	if ( (*p) == 110 )
 		goto st6;
 	goto st0;
 st6:
 	if ( ++p == pe )
-		goto _out6;
+		goto _test_eof6;
 case 6:
 	if ( (*p) == 105 )
 		goto st7;
 	goto st0;
 st7:
 	if ( ++p == pe )
-		goto _out7;
+		goto _test_eof7;
 case 7:
 	if ( (*p) == 116 )
 		goto st8;
 	goto st0;
 st8:
 	if ( ++p == pe )
-		goto _out8;
+		goto _test_eof8;
 case 8:
 	if ( (*p) == 121 )
 		goto tr16;
 	goto st0;
 st9:
 	if ( ++p == pe )
-		goto _out9;
+		goto _test_eof9;
 case 9:
 	if ( (*p) == 97 )
 		goto st10;
 	goto st0;
 st10:
 	if ( ++p == pe )
-		goto _out10;
+		goto _test_eof10;
 case 10:
 	if ( (*p) == 78 )
 		goto tr18;
 	goto st0;
 st11:
 	if ( ++p == pe )
-		goto _out11;
+		goto _test_eof11;
 case 11:
 	if ( (*p) == 97 )
 		goto st12;
 	goto st0;
 st12:
 	if ( ++p == pe )
-		goto _out12;
+		goto _test_eof12;
 case 12:
 	if ( (*p) == 108 )
 		goto st13;
 	goto st0;
 st13:
 	if ( ++p == pe )
-		goto _out13;
+		goto _test_eof13;
 case 13:
 	if ( (*p) == 115 )
 		goto st14;
 	goto st0;
 st14:
 	if ( ++p == pe )
-		goto _out14;
+		goto _test_eof14;
 case 14:
 	if ( (*p) == 101 )
 		goto tr22;
 	goto st0;
 st15:
 	if ( ++p == pe )
-		goto _out15;
+		goto _test_eof15;
 case 15:
 	if ( (*p) == 117 )
 		goto st16;
 	goto st0;
 st16:
 	if ( ++p == pe )
-		goto _out16;
+		goto _test_eof16;
 case 16:
 	if ( (*p) == 108 )
 		goto st17;
 	goto st0;
 st17:
 	if ( ++p == pe )
-		goto _out17;
+		goto _test_eof17;
 case 17:
 	if ( (*p) == 108 )
 		goto tr25;
 	goto st0;
 st18:
 	if ( ++p == pe )
-		goto _out18;
+		goto _test_eof18;
 case 18:
 	if ( (*p) == 114 )
 		goto st19;
 	goto st0;
 st19:
 	if ( ++p == pe )
-		goto _out19;
+		goto _test_eof19;
 case 19:
 	if ( (*p) == 117 )
 		goto st20;
 	goto st0;
 st20:
 	if ( ++p == pe )
-		goto _out20;
+		goto _test_eof20;
 case 20:
 	if ( (*p) == 101 )
 		goto tr28;
 	goto st0;
 	}
-	_out0: cs = 0; goto _out; 
-	_out21: cs = 21; goto _out; 
-	_out2: cs = 2; goto _out; 
-	_out3: cs = 3; goto _out; 
-	_out4: cs = 4; goto _out; 
-	_out5: cs = 5; goto _out; 
-	_out6: cs = 6; goto _out; 
-	_out7: cs = 7; goto _out; 
-	_out8: cs = 8; goto _out; 
-	_out9: cs = 9; goto _out; 
-	_out10: cs = 10; goto _out; 
-	_out11: cs = 11; goto _out; 
-	_out12: cs = 12; goto _out; 
-	_out13: cs = 13; goto _out; 
-	_out14: cs = 14; goto _out; 
-	_out15: cs = 15; goto _out; 
-	_out16: cs = 16; goto _out; 
-	_out17: cs = 17; goto _out; 
-	_out18: cs = 18; goto _out; 
-	_out19: cs = 19; goto _out; 
-	_out20: cs = 20; goto _out; 
+	_test_eof21: cs = 21; goto _test_eof; 
+	_test_eof2: cs = 2; goto _test_eof; 
+	_test_eof3: cs = 3; goto _test_eof; 
+	_test_eof4: cs = 4; goto _test_eof; 
+	_test_eof5: cs = 5; goto _test_eof; 
+	_test_eof6: cs = 6; goto _test_eof; 
+	_test_eof7: cs = 7; goto _test_eof; 
+	_test_eof8: cs = 8; goto _test_eof; 
+	_test_eof9: cs = 9; goto _test_eof; 
+	_test_eof10: cs = 10; goto _test_eof; 
+	_test_eof11: cs = 11; goto _test_eof; 
+	_test_eof12: cs = 12; goto _test_eof; 
+	_test_eof13: cs = 13; goto _test_eof; 
+	_test_eof14: cs = 14; goto _test_eof; 
+	_test_eof15: cs = 15; goto _test_eof; 
+	_test_eof16: cs = 16; goto _test_eof; 
+	_test_eof17: cs = 17; goto _test_eof; 
+	_test_eof18: cs = 18; goto _test_eof; 
+	_test_eof19: cs = 19; goto _test_eof; 
+	_test_eof20: cs = 20; goto _test_eof; 
 
+	_test_eof: {}
 	_out: {}
 	}
-#line 220 "parser.rl"
+
+#line 239 "parser.rl"
 
     if (cs >= JSON_value_first_final) {
         return p;
@@ -707,14 +736,15 @@ case 20:
 }
 
 
-#line 711 "parser.c"
+#line 740 "parser.c"
 static const int JSON_integer_start = 1;
 static const int JSON_integer_first_final = 5;
 static const int JSON_integer_error = 0;
 
 static const int JSON_integer_en_main = 1;
 
-#line 236 "parser.rl"
+
+#line 255 "parser.rl"
 
 
 static char *JSON_parse_integer(JSON_Parser *json, char *p, char *pe, VALUE *result)
@@ -722,17 +752,18 @@ static char *JSON_parse_integer(JSON_Parser *json, char *p, char *pe, VALUE *res
     int cs = EVIL;
 
     
-#line 726 "parser.c"
+#line 756 "parser.c"
 	{
 	cs = JSON_integer_start;
 	}
-#line 243 "parser.rl"
+
+#line 262 "parser.rl"
     json->memo = p;
     
-#line 733 "parser.c"
+#line 764 "parser.c"
 	{
 	if ( p == pe )
-		goto _out;
+		goto _test_eof;
 	switch ( cs )
 	{
 case 1:
@@ -744,10 +775,11 @@ case 1:
 		goto st4;
 	goto st0;
 st0:
-	goto _out0;
+cs = 0;
+	goto _out;
 st2:
 	if ( ++p == pe )
-		goto _out2;
+		goto _test_eof2;
 case 2:
 	if ( (*p) == 48 )
 		goto st3;
@@ -756,38 +788,39 @@ case 2:
 	goto st0;
 st3:
 	if ( ++p == pe )
-		goto _out3;
+		goto _test_eof3;
 case 3:
 	if ( 48 <= (*p) && (*p) <= 57 )
 		goto st0;
 	goto tr4;
 tr4:
-#line 233 "parser.rl"
-	{ goto _out5; }
+#line 252 "parser.rl"
+	{ p--; {p++; cs = 5; goto _out;} }
 	goto st5;
 st5:
 	if ( ++p == pe )
-		goto _out5;
+		goto _test_eof5;
 case 5:
-#line 773 "parser.c"
+#line 805 "parser.c"
 	goto st0;
 st4:
 	if ( ++p == pe )
-		goto _out4;
+		goto _test_eof4;
 case 4:
 	if ( 48 <= (*p) && (*p) <= 57 )
 		goto st4;
 	goto tr4;
 	}
-	_out0: cs = 0; goto _out; 
-	_out2: cs = 2; goto _out; 
-	_out3: cs = 3; goto _out; 
-	_out5: cs = 5; goto _out; 
-	_out4: cs = 4; goto _out; 
+	_test_eof2: cs = 2; goto _test_eof; 
+	_test_eof3: cs = 3; goto _test_eof; 
+	_test_eof5: cs = 5; goto _test_eof; 
+	_test_eof4: cs = 4; goto _test_eof; 
 
+	_test_eof: {}
 	_out: {}
 	}
-#line 245 "parser.rl"
+
+#line 264 "parser.rl"
 
     if (cs >= JSON_integer_first_final) {
         long len = p - json->memo;
@@ -799,14 +832,15 @@ case 4:
 }
 
 
-#line 803 "parser.c"
+#line 836 "parser.c"
 static const int JSON_float_start = 1;
 static const int JSON_float_first_final = 10;
 static const int JSON_float_error = 0;
 
 static const int JSON_float_en_main = 1;
 
-#line 267 "parser.rl"
+
+#line 286 "parser.rl"
 
 
 static char *JSON_parse_float(JSON_Parser *json, char *p, char *pe, VALUE *result)
@@ -814,17 +848,18 @@ static char *JSON_parse_float(JSON_Parser *json, char *p, char *pe, VALUE *resul
     int cs = EVIL;
 
     
-#line 818 "parser.c"
+#line 852 "parser.c"
 	{
 	cs = JSON_float_start;
 	}
-#line 274 "parser.rl"
+
+#line 293 "parser.rl"
     json->memo = p;
     
-#line 825 "parser.c"
+#line 860 "parser.c"
 	{
 	if ( p == pe )
-		goto _out;
+		goto _test_eof;
 	switch ( cs )
 	{
 case 1:
@@ -836,10 +871,11 @@ case 1:
 		goto st9;
 	goto st0;
 st0:
-	goto _out0;
+cs = 0;
+	goto _out;
 st2:
 	if ( ++p == pe )
-		goto _out2;
+		goto _test_eof2;
 case 2:
 	if ( (*p) == 48 )
 		goto st3;
@@ -848,7 +884,7 @@ case 2:
 	goto st0;
 st3:
 	if ( ++p == pe )
-		goto _out3;
+		goto _test_eof3;
 case 3:
 	switch( (*p) ) {
 		case 46: goto st4;
@@ -858,14 +894,14 @@ case 3:
 	goto st0;
 st4:
 	if ( ++p == pe )
-		goto _out4;
+		goto _test_eof4;
 case 4:
 	if ( 48 <= (*p) && (*p) <= 57 )
 		goto st5;
 	goto st0;
 st5:
 	if ( ++p == pe )
-		goto _out5;
+		goto _test_eof5;
 case 5:
 	switch( (*p) ) {
 		case 69: goto st6;
@@ -878,18 +914,18 @@ case 5:
 		goto st0;
 	goto tr7;
 tr7:
-#line 261 "parser.rl"
-	{ goto _out10; }
+#line 280 "parser.rl"
+	{ p--; {p++; cs = 10; goto _out;} }
 	goto st10;
 st10:
 	if ( ++p == pe )
-		goto _out10;
+		goto _test_eof10;
 case 10:
-#line 889 "parser.c"
+#line 925 "parser.c"
 	goto st0;
 st6:
 	if ( ++p == pe )
-		goto _out6;
+		goto _test_eof6;
 case 6:
 	switch( (*p) ) {
 		case 43: goto st7;
@@ -900,14 +936,14 @@ case 6:
 	goto st0;
 st7:
 	if ( ++p == pe )
-		goto _out7;
+		goto _test_eof7;
 case 7:
 	if ( 48 <= (*p) && (*p) <= 57 )
 		goto st8;
 	goto st0;
 st8:
 	if ( ++p == pe )
-		goto _out8;
+		goto _test_eof8;
 case 8:
 	switch( (*p) ) {
 		case 69: goto st0;
@@ -921,7 +957,7 @@ case 8:
 	goto tr7;
 st9:
 	if ( ++p == pe )
-		goto _out9;
+		goto _test_eof9;
 case 9:
 	switch( (*p) ) {
 		case 46: goto st4;
@@ -932,20 +968,21 @@ case 9:
 		goto st9;
 	goto st0;
 	}
-	_out0: cs = 0; goto _out; 
-	_out2: cs = 2; goto _out; 
-	_out3: cs = 3; goto _out; 
-	_out4: cs = 4; goto _out; 
-	_out5: cs = 5; goto _out; 
-	_out10: cs = 10; goto _out; 
-	_out6: cs = 6; goto _out; 
-	_out7: cs = 7; goto _out; 
-	_out8: cs = 8; goto _out; 
-	_out9: cs = 9; goto _out; 
+	_test_eof2: cs = 2; goto _test_eof; 
+	_test_eof3: cs = 3; goto _test_eof; 
+	_test_eof4: cs = 4; goto _test_eof; 
+	_test_eof5: cs = 5; goto _test_eof; 
+	_test_eof10: cs = 10; goto _test_eof; 
+	_test_eof6: cs = 6; goto _test_eof; 
+	_test_eof7: cs = 7; goto _test_eof; 
+	_test_eof8: cs = 8; goto _test_eof; 
+	_test_eof9: cs = 9; goto _test_eof; 
 
+	_test_eof: {}
 	_out: {}
 	}
-#line 276 "parser.rl"
+
+#line 295 "parser.rl"
 
     if (cs >= JSON_float_first_final) {
         long len = p - json->memo;
@@ -958,36 +995,39 @@ case 9:
 
 
 
-#line 962 "parser.c"
+#line 999 "parser.c"
 static const int JSON_array_start = 1;
 static const int JSON_array_first_final = 17;
 static const int JSON_array_error = 0;
 
 static const int JSON_array_en_main = 1;
 
-#line 312 "parser.rl"
+
+#line 331 "parser.rl"
 
 
 static char *JSON_parse_array(JSON_Parser *json, char *p, char *pe, VALUE *result)
 {
     int cs = EVIL;
+    VALUE array_class = json->array_class;
 
     if (json->max_nesting && json->current_nesting > json->max_nesting) {
-        rb_raise(eNestingError, "nesting of %d is to deep", json->current_nesting);
+        rb_raise(eNestingError, "nesting of %d is too deep", json->current_nesting);
     }
-    *result = rb_ary_new();
+    *result = NIL_P(array_class) ? rb_ary_new() : rb_class_new_instance(0, 0, array_class);
 
     
-#line 982 "parser.c"
+#line 1021 "parser.c"
 	{
 	cs = JSON_array_start;
 	}
-#line 324 "parser.rl"
+
+#line 344 "parser.rl"
     
-#line 988 "parser.c"
+#line 1028 "parser.c"
 	{
 	if ( p == pe )
-		goto _out;
+		goto _test_eof;
 	switch ( cs )
 	{
 case 1:
@@ -995,10 +1035,11 @@ case 1:
 		goto st2;
 	goto st0;
 st0:
-	goto _out0;
+cs = 0;
+	goto _out;
 st2:
 	if ( ++p == pe )
-		goto _out2;
+		goto _test_eof2;
 case 2:
 	switch( (*p) ) {
 		case 13: goto st2;
@@ -1022,12 +1063,12 @@ case 2:
 		goto st2;
 	goto st0;
 tr2:
-#line 293 "parser.rl"
+#line 312 "parser.rl"
 	{
         VALUE v = Qnil;
         char *np = JSON_parse_value(json, p, pe, &v); 
         if (np == NULL) {
-            goto _out3;
+            p--; {p++; cs = 3; goto _out;}
         } else {
             rb_ary_push(*result, v);
             {p = (( np))-1;}
@@ -1036,9 +1077,9 @@ tr2:
 	goto st3;
 st3:
 	if ( ++p == pe )
-		goto _out3;
+		goto _test_eof3;
 case 3:
-#line 1042 "parser.c"
+#line 1083 "parser.c"
 	switch( (*p) ) {
 		case 13: goto st3;
 		case 32: goto st3;
@@ -1051,7 +1092,7 @@ case 3:
 	goto st0;
 st4:
 	if ( ++p == pe )
-		goto _out4;
+		goto _test_eof4;
 case 4:
 	switch( (*p) ) {
 		case 13: goto st4;
@@ -1075,7 +1116,7 @@ case 4:
 	goto st0;
 st5:
 	if ( ++p == pe )
-		goto _out5;
+		goto _test_eof5;
 case 5:
 	switch( (*p) ) {
 		case 42: goto st6;
@@ -1084,14 +1125,14 @@ case 5:
 	goto st0;
 st6:
 	if ( ++p == pe )
-		goto _out6;
+		goto _test_eof6;
 case 6:
 	if ( (*p) == 42 )
 		goto st7;
 	goto st6;
 st7:
 	if ( ++p == pe )
-		goto _out7;
+		goto _test_eof7;
 case 7:
 	switch( (*p) ) {
 		case 42: goto st7;
@@ -1100,14 +1141,14 @@ case 7:
 	goto st6;
 st8:
 	if ( ++p == pe )
-		goto _out8;
+		goto _test_eof8;
 case 8:
 	if ( (*p) == 10 )
 		goto st4;
 	goto st8;
 st9:
 	if ( ++p == pe )
-		goto _out9;
+		goto _test_eof9;
 case 9:
 	switch( (*p) ) {
 		case 42: goto st10;
@@ -1116,14 +1157,14 @@ case 9:
 	goto st0;
 st10:
 	if ( ++p == pe )
-		goto _out10;
+		goto _test_eof10;
 case 10:
 	if ( (*p) == 42 )
 		goto st11;
 	goto st10;
 st11:
 	if ( ++p == pe )
-		goto _out11;
+		goto _test_eof11;
 case 11:
 	switch( (*p) ) {
 		case 42: goto st11;
@@ -1132,24 +1173,24 @@ case 11:
 	goto st10;
 st12:
 	if ( ++p == pe )
-		goto _out12;
+		goto _test_eof12;
 case 12:
 	if ( (*p) == 10 )
 		goto st3;
 	goto st12;
 tr4:
-#line 304 "parser.rl"
-	{ goto _out17; }
+#line 323 "parser.rl"
+	{ p--; {p++; cs = 17; goto _out;} }
 	goto st17;
 st17:
 	if ( ++p == pe )
-		goto _out17;
+		goto _test_eof17;
 case 17:
-#line 1149 "parser.c"
+#line 1190 "parser.c"
 	goto st0;
 st13:
 	if ( ++p == pe )
-		goto _out13;
+		goto _test_eof13;
 case 13:
 	switch( (*p) ) {
 		case 42: goto st14;
@@ -1158,14 +1199,14 @@ case 13:
 	goto st0;
 st14:
 	if ( ++p == pe )
-		goto _out14;
+		goto _test_eof14;
 case 14:
 	if ( (*p) == 42 )
 		goto st15;
 	goto st14;
 st15:
 	if ( ++p == pe )
-		goto _out15;
+		goto _test_eof15;
 case 15:
 	switch( (*p) ) {
 		case 42: goto st15;
@@ -1174,38 +1215,39 @@ case 15:
 	goto st14;
 st16:
 	if ( ++p == pe )
-		goto _out16;
+		goto _test_eof16;
 case 16:
 	if ( (*p) == 10 )
 		goto st2;
 	goto st16;
 	}
-	_out0: cs = 0; goto _out; 
-	_out2: cs = 2; goto _out; 
-	_out3: cs = 3; goto _out; 
-	_out4: cs = 4; goto _out; 
-	_out5: cs = 5; goto _out; 
-	_out6: cs = 6; goto _out; 
-	_out7: cs = 7; goto _out; 
-	_out8: cs = 8; goto _out; 
-	_out9: cs = 9; goto _out; 
-	_out10: cs = 10; goto _out; 
-	_out11: cs = 11; goto _out; 
-	_out12: cs = 12; goto _out; 
-	_out17: cs = 17; goto _out; 
-	_out13: cs = 13; goto _out; 
-	_out14: cs = 14; goto _out; 
-	_out15: cs = 15; goto _out; 
-	_out16: cs = 16; goto _out; 
+	_test_eof2: cs = 2; goto _test_eof; 
+	_test_eof3: cs = 3; goto _test_eof; 
+	_test_eof4: cs = 4; goto _test_eof; 
+	_test_eof5: cs = 5; goto _test_eof; 
+	_test_eof6: cs = 6; goto _test_eof; 
+	_test_eof7: cs = 7; goto _test_eof; 
+	_test_eof8: cs = 8; goto _test_eof; 
+	_test_eof9: cs = 9; goto _test_eof; 
+	_test_eof10: cs = 10; goto _test_eof; 
+	_test_eof11: cs = 11; goto _test_eof; 
+	_test_eof12: cs = 12; goto _test_eof; 
+	_test_eof17: cs = 17; goto _test_eof; 
+	_test_eof13: cs = 13; goto _test_eof; 
+	_test_eof14: cs = 14; goto _test_eof; 
+	_test_eof15: cs = 15; goto _test_eof; 
+	_test_eof16: cs = 16; goto _test_eof; 
 
+	_test_eof: {}
 	_out: {}
 	}
-#line 325 "parser.rl"
+
+#line 345 "parser.rl"
 
     if(cs >= JSON_array_first_final) {
         return p + 1;
     } else {
-        rb_raise(eParserError, "unexpected token at '%s'", p);
+        rb_raise(eParserError, "%u: unexpected token at '%s'", __LINE__, p);
     }
 }
 
@@ -1266,14 +1308,15 @@ static VALUE json_string_unescape(char *p, char *pe)
 }
 
 
-#line 1270 "parser.c"
+#line 1312 "parser.c"
 static const int JSON_string_start = 1;
 static const int JSON_string_first_final = 8;
 static const int JSON_string_error = 0;
 
 static const int JSON_string_en_main = 1;
 
-#line 403 "parser.rl"
+
+#line 429 "parser.rl"
 
 
 static char *JSON_parse_string(JSON_Parser *json, char *p, char *pe, VALUE *result)
@@ -1282,17 +1325,18 @@ static char *JSON_parse_string(JSON_Parser *json, char *p, char *pe, VALUE *resu
 
     *result = rb_str_new("", 0);
     
-#line 1286 "parser.c"
+#line 1329 "parser.c"
 	{
 	cs = JSON_string_start;
 	}
-#line 411 "parser.rl"
+
+#line 437 "parser.rl"
     json->memo = p;
     
-#line 1293 "parser.c"
+#line 1337 "parser.c"
 	{
 	if ( p == pe )
-		goto _out;
+		goto _test_eof;
 	switch ( cs )
 	{
 case 1:
@@ -1300,10 +1344,11 @@ case 1:
 		goto st2;
 	goto st0;
 st0:
-	goto _out0;
+cs = 0;
+	goto _out;
 st2:
 	if ( ++p == pe )
-		goto _out2;
+		goto _test_eof2;
 case 2:
 	switch( (*p) ) {
 		case 34: goto tr2;
@@ -1313,23 +1358,29 @@ case 2:
 		goto st0;
 	goto st2;
 tr2:
-#line 395 "parser.rl"
+#line 415 "parser.rl"
 	{
         *result = json_string_unescape(json->memo + 1, p);
-        if (NIL_P(*result)) goto _out8; else {p = (( p + 1))-1;}
-    }
-#line 400 "parser.rl"
-	{ goto _out8; }
+        if (NIL_P(*result)) {
+			p--;
+			{p++; cs = 8; goto _out;}
+		} else {
+			FORCE_UTF8(*result);
+			{p = (( p + 1))-1;}
+		}
+	}
+#line 426 "parser.rl"
+	{ p--; {p++; cs = 8; goto _out;} }
 	goto st8;
 st8:
 	if ( ++p == pe )
-		goto _out8;
+		goto _test_eof8;
 case 8:
-#line 1329 "parser.c"
+#line 1380 "parser.c"
 	goto st0;
 st3:
 	if ( ++p == pe )
-		goto _out3;
+		goto _test_eof3;
 case 3:
 	if ( (*p) == 117 )
 		goto st4;
@@ -1338,7 +1389,7 @@ case 3:
 	goto st2;
 st4:
 	if ( ++p == pe )
-		goto _out4;
+		goto _test_eof4;
 case 4:
 	if ( (*p) < 65 ) {
 		if ( 48 <= (*p) && (*p) <= 57 )
@@ -1351,7 +1402,7 @@ case 4:
 	goto st0;
 st5:
 	if ( ++p == pe )
-		goto _out5;
+		goto _test_eof5;
 case 5:
 	if ( (*p) < 65 ) {
 		if ( 48 <= (*p) && (*p) <= 57 )
@@ -1364,7 +1415,7 @@ case 5:
 	goto st0;
 st6:
 	if ( ++p == pe )
-		goto _out6;
+		goto _test_eof6;
 case 6:
 	if ( (*p) < 65 ) {
 		if ( 48 <= (*p) && (*p) <= 57 )
@@ -1377,7 +1428,7 @@ case 6:
 	goto st0;
 st7:
 	if ( ++p == pe )
-		goto _out7;
+		goto _test_eof7;
 case 7:
 	if ( (*p) < 65 ) {
 		if ( 48 <= (*p) && (*p) <= 57 )
@@ -1389,18 +1440,19 @@ case 7:
 		goto st2;
 	goto st0;
 	}
-	_out0: cs = 0; goto _out; 
-	_out2: cs = 2; goto _out; 
-	_out8: cs = 8; goto _out; 
-	_out3: cs = 3; goto _out; 
-	_out4: cs = 4; goto _out; 
-	_out5: cs = 5; goto _out; 
-	_out6: cs = 6; goto _out; 
-	_out7: cs = 7; goto _out; 
+	_test_eof2: cs = 2; goto _test_eof; 
+	_test_eof8: cs = 8; goto _test_eof; 
+	_test_eof3: cs = 3; goto _test_eof; 
+	_test_eof4: cs = 4; goto _test_eof; 
+	_test_eof5: cs = 5; goto _test_eof; 
+	_test_eof6: cs = 6; goto _test_eof; 
+	_test_eof7: cs = 7; goto _test_eof; 
 
+	_test_eof: {}
 	_out: {}
 	}
-#line 413 "parser.rl"
+
+#line 439 "parser.rl"
 
     if (cs >= JSON_string_first_final) {
         return p + 1;
@@ -1411,14 +1463,15 @@ case 7:
 
 
 
-#line 1415 "parser.c"
+#line 1467 "parser.c"
 static const int JSON_start = 1;
 static const int JSON_first_final = 10;
 static const int JSON_error = 0;
 
 static const int JSON_en_main = 1;
 
-#line 447 "parser.rl"
+
+#line 473 "parser.rl"
 
 
 /* 
@@ -1453,6 +1506,8 @@ static const int JSON_en_main = 1;
  * * *create_additions*: If set to false, the Parser doesn't create
  *   additions even if a matchin class and create_id was found. This option
  *   defaults to true.
+ * * *object_class*: Defaults to Hash
+ * * *array_class*: Defaults to Array
  */
 static VALUE cParser_initialize(int argc, VALUE *argv, VALUE self)
 {
@@ -1473,7 +1528,7 @@ static VALUE cParser_initialize(int argc, VALUE *argv, VALUE self)
             rb_raise(rb_eArgError, "opts needs to be like a hash");
         } else {
             VALUE tmp = ID2SYM(i_max_nesting);
-            if (st_lookup(RHASH(opts)->tbl, tmp, 0)) {
+            if (st_lookup(RHASH_TBL(opts), tmp, 0)) {
                 VALUE max_nesting = rb_hash_aref(opts, tmp);
                 if (RTEST(max_nesting)) {
                     Check_Type(max_nesting, T_FIXNUM);
@@ -1485,14 +1540,14 @@ static VALUE cParser_initialize(int argc, VALUE *argv, VALUE self)
                 json->max_nesting = 19;
             }
             tmp = ID2SYM(i_allow_nan);
-            if (st_lookup(RHASH(opts)->tbl, tmp, 0)) {
+            if (st_lookup(RHASH_TBL(opts), tmp, 0)) {
                 VALUE allow_nan = rb_hash_aref(opts, tmp);
                 json->allow_nan = RTEST(allow_nan) ? 1 : 0;
             } else {
                 json->allow_nan = 0;
             }
             tmp = ID2SYM(i_create_additions);
-            if (st_lookup(RHASH(opts)->tbl, tmp, 0)) {
+            if (st_lookup(RHASH_TBL(opts), tmp, 0)) {
                 VALUE create_additions = rb_hash_aref(opts, tmp);
                 if (RTEST(create_additions)) {
                     json->create_id = rb_funcall(mJSON, i_create_id, 0);
@@ -1502,11 +1557,25 @@ static VALUE cParser_initialize(int argc, VALUE *argv, VALUE self)
             } else {
                 json->create_id = rb_funcall(mJSON, i_create_id, 0);
             }
+            tmp = ID2SYM(i_object_class);
+            if (st_lookup(RHASH_TBL(opts), tmp, 0)) {
+                json->object_class = rb_hash_aref(opts, tmp);
+            } else {
+                json->object_class = Qnil;
+            }
+            tmp = ID2SYM(i_array_class);
+            if (st_lookup(RHASH_TBL(opts), tmp, 0)) {
+                json->array_class = rb_hash_aref(opts, tmp);
+            } else {
+                json->array_class = Qnil;
+            }
         }
     } else {
         json->max_nesting = 19;
         json->allow_nan = 0;
         json->create_id = rb_funcall(mJSON, i_create_id, 0);
+        json->object_class = Qnil;
+        json->array_class = Qnil;
     }
     json->current_nesting = 0;
     /*
@@ -1541,23 +1610,24 @@ static VALUE cParser_parse(VALUE self)
     GET_STRUCT;
 
     
-#line 1545 "parser.c"
+#line 1614 "parser.c"
 	{
 	cs = JSON_start;
 	}
-#line 569 "parser.rl"
+
+#line 611 "parser.rl"
     p = json->source;
     pe = p + json->len;
     
-#line 1553 "parser.c"
+#line 1623 "parser.c"
 	{
 	if ( p == pe )
-		goto _out;
+		goto _test_eof;
 	switch ( cs )
 	{
 st1:
 	if ( ++p == pe )
-		goto _out1;
+		goto _test_eof1;
 case 1:
 	switch( (*p) ) {
 		case 13: goto st1;
@@ -1570,10 +1640,11 @@ case 1:
 		goto st1;
 	goto st0;
 st0:
-	goto _out0;
+cs = 0;
+	goto _out;
 st2:
 	if ( ++p == pe )
-		goto _out2;
+		goto _test_eof2;
 case 2:
 	switch( (*p) ) {
 		case 42: goto st3;
@@ -1582,14 +1653,14 @@ case 2:
 	goto st0;
 st3:
 	if ( ++p == pe )
-		goto _out3;
+		goto _test_eof3;
 case 3:
 	if ( (*p) == 42 )
 		goto st4;
 	goto st3;
 st4:
 	if ( ++p == pe )
-		goto _out4;
+		goto _test_eof4;
 case 4:
 	switch( (*p) ) {
 		case 42: goto st4;
@@ -1598,34 +1669,34 @@ case 4:
 	goto st3;
 st5:
 	if ( ++p == pe )
-		goto _out5;
+		goto _test_eof5;
 case 5:
 	if ( (*p) == 10 )
 		goto st1;
 	goto st5;
 tr3:
-#line 436 "parser.rl"
+#line 462 "parser.rl"
 	{
         char *np;
         json->current_nesting = 1;
         np = JSON_parse_array(json, p, pe, &result);
-        if (np == NULL) goto _out10; else {p = (( np))-1;}
+        if (np == NULL) { p--; {p++; cs = 10; goto _out;} } else {p = (( np))-1;}
     }
 	goto st10;
 tr4:
-#line 429 "parser.rl"
+#line 455 "parser.rl"
 	{
         char *np;
         json->current_nesting = 1;
         np = JSON_parse_object(json, p, pe, &result);
-        if (np == NULL) goto _out10; else {p = (( np))-1;}
+        if (np == NULL) { p--; {p++; cs = 10; goto _out;} } else {p = (( np))-1;}
     }
 	goto st10;
 st10:
 	if ( ++p == pe )
-		goto _out10;
+		goto _test_eof10;
 case 10:
-#line 1629 "parser.c"
+#line 1700 "parser.c"
 	switch( (*p) ) {
 		case 13: goto st10;
 		case 32: goto st10;
@@ -1636,7 +1707,7 @@ case 10:
 	goto st0;
 st6:
 	if ( ++p == pe )
-		goto _out6;
+		goto _test_eof6;
 case 6:
 	switch( (*p) ) {
 		case 42: goto st7;
@@ -1645,14 +1716,14 @@ case 6:
 	goto st0;
 st7:
 	if ( ++p == pe )
-		goto _out7;
+		goto _test_eof7;
 case 7:
 	if ( (*p) == 42 )
 		goto st8;
 	goto st7;
 st8:
 	if ( ++p == pe )
-		goto _out8;
+		goto _test_eof8;
 case 8:
 	switch( (*p) ) {
 		case 42: goto st8;
@@ -1661,32 +1732,33 @@ case 8:
 	goto st7;
 st9:
 	if ( ++p == pe )
-		goto _out9;
+		goto _test_eof9;
 case 9:
 	if ( (*p) == 10 )
 		goto st10;
 	goto st9;
 	}
-	_out1: cs = 1; goto _out; 
-	_out0: cs = 0; goto _out; 
-	_out2: cs = 2; goto _out; 
-	_out3: cs = 3; goto _out; 
-	_out4: cs = 4; goto _out; 
-	_out5: cs = 5; goto _out; 
-	_out10: cs = 10; goto _out; 
-	_out6: cs = 6; goto _out; 
-	_out7: cs = 7; goto _out; 
-	_out8: cs = 8; goto _out; 
-	_out9: cs = 9; goto _out; 
+	_test_eof1: cs = 1; goto _test_eof; 
+	_test_eof2: cs = 2; goto _test_eof; 
+	_test_eof3: cs = 3; goto _test_eof; 
+	_test_eof4: cs = 4; goto _test_eof; 
+	_test_eof5: cs = 5; goto _test_eof; 
+	_test_eof10: cs = 10; goto _test_eof; 
+	_test_eof6: cs = 6; goto _test_eof; 
+	_test_eof7: cs = 7; goto _test_eof; 
+	_test_eof8: cs = 8; goto _test_eof; 
+	_test_eof9: cs = 9; goto _test_eof; 
 
+	_test_eof: {}
 	_out: {}
 	}
-#line 572 "parser.rl"
+
+#line 614 "parser.rl"
 
     if (cs >= JSON_first_final && p == pe) {
         return result;
     } else {
-        rb_raise(eParserError, "unexpected token at '%s'", p);
+        rb_raise(eParserError, "%u: unexpected token at '%s'", __LINE__, p);
     }
 }
 
@@ -1701,11 +1773,13 @@ static void JSON_mark(JSON_Parser *json)
 {
     rb_gc_mark_maybe(json->Vsource);
     rb_gc_mark_maybe(json->create_id);
+    rb_gc_mark_maybe(json->object_class);
+    rb_gc_mark_maybe(json->array_class);
 }
 
 static void JSON_free(JSON_Parser *json)
 {
-    free(json);
+    ruby_xfree(json);
 }
 
 static VALUE cJSON_parser_s_allocate(VALUE klass)
@@ -1728,6 +1802,7 @@ static VALUE cParser_source(VALUE self)
 
 void Init_parser()
 {
+    rb_require("json/common");
     mJSON = rb_define_module("JSON");
     mExt = rb_define_module_under(mJSON, "Ext");
     cParser = rb_define_class_under(mExt, "Parser", rb_cObject);
@@ -1749,4 +1824,6 @@ void Init_parser()
     i_chr = rb_intern("chr");
     i_max_nesting = rb_intern("max_nesting");
     i_allow_nan = rb_intern("allow_nan");
+    i_object_class = rb_intern("object_class");
+    i_array_class = rb_intern("array_class");
 }
